import pytest

from tests.utils import build_rule


@pytest.mark.parametrize(
    "event,expected",
    [
        ({"a": "foo"}, False),
        ({"a": "foo", "b": "bar"}, False),
        ({"b": "foo"}, False),
        ({"d": "foo"}, False),
        ({"c": "foo"}, True),
        ({"a": "foo", "c": "foo"}, True),
    ],
)
def test_exists(event: dict, expected: bool):
    rule = build_rule("""
detection:
  foo:
    a|exists: true
  bar:
    b|exists: false
  baz:
    c|exists: true
  condition: (foo or bar) and baz
""")
    assert rule.match(event) is expected



import pytest

from tests.utils import build_rule


@pytest.mark.parametrize(
    "event,expected",
    [
        ({"Foo": "bar"}, True),
        ({"foo": "bar"}, False),
        ({"a": "bar", "b": "foo"}, False),
    ],
)
def test_selection(event: dict, expected: bool):
    rule = build_rule("""
detection:
  selection:
    Foo: bar
  condition: selection
""")
    assert rule.match(event) is expected



import pytest

from tests.utils import build_rule


@pytest.mark.parametrize(
    "event,expected",
    [
        ({"a": 11}, False),
        ({"a": 10}, True),
        ({"a": 9}, True),
    ],
)
def test_lte(event: dict, expected: bool):
    rule = build_rule("""
detection:
  foo:
    a|lte: 10
  condition: foo
""")
    assert rule.match(event) is expected



from functools import lru_cache

import yaml

from azuma.schemas import Rule


@lru_cache
def build_rule(
    detection: str,
    title: str | None = None,
    logsource: str | None = None,
):
    loaded = yaml.safe_load(detection)
    _detection = loaded.get("detection") or loaded

    rule = {
        "title": title or "dummy",
        "logsource": logsource or {"category": "dummy"},
        "detection": _detection,
    }
    return Rule.model_validate(rule)



import base64

import pytest

from tests.utils import build_rule


@pytest.mark.parametrize(
    "event,expected",
    [
        ({"a": base64.encodebytes(b"/bin/bash").decode()}, True),
        ({"a": base64.b64encode(b"/bin/sh").decode()}, True),
        ({"a": base64.encodebytes(b"/BIN/BASH").decode()}, False),
        ({"a": base64.b64encode(b"/BIN/SH").decode()}, False),
        ({"a": "L2Jpbi9iYXNo"}, True),
        ({"a": "9iaW4vYmFza"}, True),
        ({"a": "vYmluL2Jhc2"}, True),
        ({"a": "foo"}, False),
    ],
)
def test_base64offset(event: dict, expected: bool):
    rule = build_rule("""
detection:
  foo:
    a|base64offset:
      - /bin/bash
      - /bin/sh
      - /bin/zsh
  condition: foo""")
    assert rule.match(event) is expected



import pytest

from tests.utils import build_rule


@pytest.mark.parametrize(
    "event,expected",
    [
        ({"foo": "bar"}, True),
        ({"foo": "Bar"}, True),
        ({"foo": "BAR"}, True),
        ({"foo": "baz"}, False),
    ],
)
def test_match(event: dict, expected: bool):
    rule = build_rule(
        """
detection:
  foo:
    - bar
  condition: foo
"""
    )
    assert rule.match(event) is expected



import contextlib
import datetime


def is_valid_date_format(v: str) -> bool:
    with contextlib.suppress(Exception):
        datetime.datetime.strptime(v, "%Y-%m-%d")
        return True

    return False



import pytest

from tests.utils import build_rule


@pytest.mark.parametrize(
    "event,expected",
    [
        ({"a": "foobar"}, True),
        ({"a": "foo"}, False),
        ({"a": "bar"}, False),
    ],
)
def test_all(event: dict, expected: bool):
    rule = build_rule("""
detection:
  foo:
    a|contains|all:
      - foo
      - bar
  condition: foo
    """)
    assert rule.match(event) is expected



import itertools
import json
import sys
from functools import partial
from typing import Annotated, cast

import senkawa
import typer
from pydantic import BaseModel, Field, ValidationError
from pydantic_core import InitErrorDetails
from returns.functions import raise_exception
from returns.pipeline import flow, is_successful
from returns.pointfree import bind
from returns.result import ResultE, safe

from azuma import __version__, schemas

app = typer.Typer()


class ScanResult(BaseModel):
    path: str = Field(...)
    matched: bool = Field(...)


@safe(exceptions=(ValidationError,))
def load_rule(path: str) -> schemas.Rule:
    return schemas.Rule.model_validate_file(path)


@app.command()
def scan(
    path: Annotated[
        str, typer.Argument(help="Path (or glob pattern) to rule YAML file(s)")
    ],
    target: Annotated[
        str, typer.Argument(help="Path (or glob pattern) to event JSON file(s)")
    ],
):
    @safe
    def load_rules(path: str) -> list[schemas.Rule]:
        results = [load_rule(path_) for path_ in senkawa.glob(path)]
        return [result.alt(raise_exception).unwrap() for result in results]

    @safe
    def scan(rules: list[schemas.Rule], *, target: str) -> dict[str, list[dict]]:
        memo: dict[str, list[dict]] = {}
        for rule in rules:
            results: list[ScanResult] = []

            for path in senkawa.glob(target):
                with open(path) as f:
                    data = json.loads(f.read())
                    results.append(ScanResult(path=path, matched=rule.match(data)))

            key = rule.id or rule.title
            memo[key] = [r.model_dump() for r in results]

        return memo

    @safe
    def output(results: dict[str, list[dict]]) -> None:
        print(json.dumps(results))  # noqa: T201

    result: ResultE[None] = flow(
        path, load_rules, bind(partial(scan, target=target)), bind(output)
    )
    result.alt(raise_exception)


OPTIONAL_FIELDS = [
    "author",
    "date",
    "description",
    "falsepositives",
    "fields",
    "id",
    "level",
    "license",
    "modified",
    "references",
    "related",
    "status",
    "tags",
]


@safe
def extra_checks(  # noqa: C901
    rule: schemas.Rule,
    *,
    check_id: bool = False,
    check_license: bool = False,
    check_author: bool = False,
    check_date: bool = False,
    check_modified: bool = False,
    check_description: bool = False,
    check_status: bool = False,
    check_level: bool = False,
    check_references: bool = False,
    check_tags: bool = False,
    check_fields: bool = False,
    check_related: bool = False,
    check_falsepositives: bool = False,
    check_all: bool = False,
) -> schemas.Rule:
    errors: list[InitErrorDetails] = []

    fields_to_check: set[str] = set(OPTIONAL_FIELDS) if check_all else set()

    if check_id:
        fields_to_check.add("id")

    if check_license:
        fields_to_check.add("license")

    if check_author:
        fields_to_check.add("author")

    if check_date:
        fields_to_check.add("date")

    if check_modified:
        fields_to_check.add("modified")

    if check_description:
        fields_to_check.add("description")

    if check_status:
        fields_to_check.add("status")

    if check_level:
        fields_to_check.add("level")

    if check_references:
        fields_to_check.add("references")

    if check_tags:
        fields_to_check.add("tags")

    if check_fields:
        fields_to_check.add("fields")

    if check_related:
        fields_to_check.add("related")

    if check_falsepositives:
        fields_to_check.add("falsepositives")

    dumped = rule.model_dump()
    for check in fields_to_check:
        if check not in dumped or dumped[check] is None:
            errors.append(InitErrorDetails(type="missing", loc=(check,)))  # type: ignore

    if len(errors) > 0:
        raise ValidationError.from_exception_data("Field required", errors)

    return rule


@app.command()
def validate(
    path: Annotated[
        list[str],
        typer.Argument(help="Path(s) (or glob pattern(s)) to rule YAML file(s)"),
    ],
    check_id: Annotated[
        bool, typer.Option(help="Check for missing 'id' field")
    ] = False,
    check_license: Annotated[
        bool, typer.Option(help="Check for missing 'license' field")
    ] = False,
    check_author: Annotated[
        bool, typer.Option(help="Check for missing 'author' field")
    ] = False,
    check_date: Annotated[
        bool, typer.Option(help="Check for missing 'date' field")
    ] = False,
    check_modified: Annotated[
        bool, typer.Option(help="Check for missing 'modified' field")
    ] = False,
    check_description: Annotated[
        bool, typer.Option(help="Check for missing 'description' field")
    ] = False,
    check_status: Annotated[
        bool, typer.Option(help="Check for missing 'status' field")
    ] = False,
    check_level: Annotated[
        bool, typer.Option(help="Check for missing 'level' field")
    ] = False,
    check_references: Annotated[
        bool, typer.Option(help="Check for missing 'references' field")
    ] = False,
    check_tags: Annotated[
        bool, typer.Option(help="Check for missing 'tags' field")
    ] = False,
    check_falsepositives: Annotated[
        bool, typer.Option(help="Check for missing 'falsepositives' field")
    ] = False,
    check_fields: Annotated[
        bool, typer.Option(help="Check for missing 'fields' field")
    ] = False,
    check_related: Annotated[
        bool, typer.Option(help="Check for missing 'related' field")
    ] = False,
    check_all: Annotated[
        bool, typer.Option(help="Check for all the missing optional fields")
    ] = False,
):
    memo: dict[str, ValidationError] = {}
    for path_ in itertools.chain.from_iterable([senkawa.glob(p) for p in path]):
        result: ResultE[schemas.Rule] = flow(
            path_,
            load_rule,
            bind(
                partial(
                    extra_checks,
                    check_id=check_id,
                    check_license=check_license,
                    check_author=check_author,
                    check_date=check_date,
                    check_modified=check_modified,
                    check_description=check_description,
                    check_status=check_status,
                    check_level=check_level,
                    check_references=check_references,
                    check_tags=check_tags,
                    check_fields=check_fields,
                    check_related=check_related,
                    check_falsepositives=check_falsepositives,
                    check_all=check_all,
                )
            ),
        )
        if not is_successful(result):
            memo[path_] = cast(ValidationError, result.failure())

    if not memo:
        return

    for path_, error in memo.items():
        print(f"{path_} has {error}\n", file=sys.stderr)  # noqa: T201

    sys.exit(1)


@app.command()
def version():
    print(__version__)  # noqa: T201


if __name__ == "__main__":
    typer.run(app())



import pytest

from tests.utils import build_rule


@pytest.mark.parametrize(
    "event,expected",
    [
        ({"log": "a"}, False),
        ({"log": "b"}, False),
        ({"log": "ab"}, True),
        ({"log": "bac"}, True),
        ({"log": "c"}, False),
    ],
)
def test_all_of_them(event: dict, expected: bool):
    rule = build_rule("""
detection:
    a: ["a"]
    b: ["b"]
    condition: all of them
""")
    assert rule.match(event) is expected


@pytest.mark.parametrize(
    "event,expected",
    [
        ({"log": "aa"}, False),
        ({"log": "1ab ba ca"}, False),
        ({"log": "ba"}, False),
        ({"log": "aabb"}, True),
    ],
)
def test_all_of_x(event: dict, expected: bool):
    rule = build_rule("""
detection:
  aa: ["aa"]
  ab: ["ab"]
  ba: ["ba"]
  bb: ["bb"]
  condition: all of a*
""")
    assert rule.match(event) is expected



import importlib.metadata

import packaging.version
from pydantic import VERSION

from .schemas import Rule, RuleSet  # noqa: F401

try:
    __version__ = importlib.metadata.version(__name__)
except importlib.metadata.PackageNotFoundError:
    __version__ = "0.0.0"


PYDANTIC_VERSION = packaging.version.parse(VERSION)

if packaging.version.parse("2.10.0") <= PYDANTIC_VERSION:
    Rule.model_rebuild()



import pytest

from tests.utils import build_rule


@pytest.mark.parametrize(
    "event,expected",
    [
        ({"a": "aa\nbb"}, True),
        ({"a": "foo\naa"}, True),
        ({"a": "bar\naa"}, True),
        ({"a": "a\na"}, False),
        ({"a": "b\nb"}, False),
    ],
)
def test_with_multi_line(event: dict, expected: bool):
    rule = build_rule("""
detection:
  foo:
    - aa
    - bb
  condition: foo
""")
    assert rule.match(event) is expected



import pytest

from tests.utils import build_rule


@pytest.mark.parametrize(
    "event,expected",
    [
        ({"foo": {"bar": "a"}}, True),
        ({"foo": {"bar": "b"}}, False),
        ({"foo": "bar"}, False),
        ({"log": "a"}, False),
    ],
)
def test_nested_dict(event: dict, expected: bool):
    rule = build_rule("""
detection:
  field:
    foo:
      bar: a
  condition: field
""")
    assert rule.match(event) is expected


@pytest.mark.parametrize(
    "event,expected",
    [
        ({"foo": {"bar": "a"}}, True),
        ({"foo": {"bar1": "a"}}, True),
        ({"foo": {"bar2": "a"}}, True),
        ({"foo": {"bar": "b"}}, False),
        ({"foo": "bar"}, False),
        ({"log": "a"}, False),
    ],
)
def test_nested_dict_with_wildcard(event: dict, expected: bool):
    rule = build_rule("""
detection:
  field:
    foo:
      bar*: a
  condition: field
""")
    assert rule.match(event) is expected



import pytest

from tests.utils import build_rule


@pytest.mark.parametrize(
    "event,expected",
    [
        ({"a": "192.168.0.1"}, True),
        ({"a": "127.0.0.1"}, False),
    ],
)
def test_cidr(event: dict, expected: bool):
    rule = build_rule("""
detection:
  foo:
    a|cidr: 192.168.0.0/24
  condition: foo""")
    assert rule.match(event) is expected



# modified from https://github.com/SigmaHQ/pySigma/
# (https://github.com/SigmaHQ/pySigma/blob/main/sigma/types.py)
import dataclasses

import regex as re


@dataclasses.dataclass
class Placeholder:
    name: str


def replace_with_placeholder(
    s: str, regex: re.Pattern, placeholder_name: str
) -> tuple[str | Placeholder, ...]:
    result: list[str | Placeholder] = []

    matched = False
    i = 0
    for m in regex.finditer(s):
        matched = True
        _s = s[i : m.start()]
        if _s != "":
            result.append(_s)
        result.append(Placeholder(placeholder_name))
        i = m.end()

    if matched:  # if matched, append remainder of string
        _s = s[i:]
        if _s != "":
            result.append(_s)
    else:  # no matches: append original string
        result.append(s)

    return tuple(result)


def replace_placeholders(
    replaced: tuple[str | Placeholder, ...], placeholder: str
) -> str:
    result: list[str] = []

    for rep in replaced:
        if isinstance(rep, Placeholder):
            result.append(placeholder)
        else:
            result.append(rep)

    return "".join(result)



from pathlib import Path

import pytest

from azuma import schemas


@pytest.fixture
def rule_set():
    return schemas.RuleSet.from_dir("tests/fixtures")


@pytest.fixture
def paths():
    p = Path("tests/fixtures")
    return list(p.glob("**/*.yml"))


def test_rules(rule_set: schemas.RuleSet, paths: list[Path]):
    assert len(rule_set.root) == len(paths)


@pytest.mark.parametrize(
    "event,expected",
    [
        ({"a": "foo", "b": "bar", "c": "baz"}, 1),
        ({"a": "foo"}, 1),
        ({"a": "foobar"}, 1),
        ({"b": "bar"}, 0),
        ({"a": "bar", "b": "foo"}, 0),
    ],
)
def test_match_all(rule_set: schemas.RuleSet, event: dict, expected: int):
    matched = rule_set.match_all(event)
    assert len(matched) == expected


def test_unique(rule_set: schemas.RuleSet):
    combined = schemas.RuleSet(root=rule_set.root + rule_set.root)
    unique = combined.unique()
    assert len(combined.root) != len(unique.root)
    assert len(rule_set.root) == len(unique.root)



from typing import Any

import pytest
import regex as re
from returns.result import safe

from azuma.parsers.detection import (
    apply_base64offset_modifier,
    apply_utf16_modifier,
    sigma_string_to_regex,
    validate_base64_sub_modifier_condition,
    validate_exists_modifier_condition,
    windash_generator,
)


@pytest.mark.parametrize(
    "v,expected",
    [
        (r".", r"\."),
        (r"*", r".*"),
        (r"?", r"."),
        (r".\*", r"\.\*"),
        (r".\?", r"\.\?"),
        (r".\*abc", r"\.\*abc"),
        (r".\*abc*", r"\.\*abc.*"),
        (r".\*abc?", r"\.\*abc."),
        (r".\*abc\?", r"\.\*abc\?"),
        (r".\*abc\\?", r"\.\*abc\\."),
        (r".\*abc\\\?", r"\.\*abc\\\\."),
    ],
)
def test_sigma_string_to_regex(v: str, expected: str):
    assert sigma_string_to_regex(v) == expected


@pytest.mark.parametrize(
    "v,expected",
    [
        (r"a\a", r"a\a"),
        (r"a\\a", r"a\\a"),
        (r"a\*a", "a*a"),
        (r"a*a", r"a a bunch of garbage a"),
    ],
)
def test_sigma_string_to_regex_with_fullmatch(v: str, expected: str):
    assert re.compile(sigma_string_to_regex(v)).fullmatch(expected)


def test_apply_base64offset_modifier():
    assert (
        apply_base64offset_modifier("/bin/bash")
        == "(L2Jpbi9iYXNo|9iaW4vYmFza|vYmluL2Jhc2)"
    )


@pytest.mark.parametrize(
    "v,encoding,expected",
    [
        ("cmd", "utf-16le", bytes.fromhex("63 00 6d 00 64 00")),
        ("cmd", "utf-16be", bytes.fromhex("00 63 00 6d 00 64")),
        ("cmd", "utf-16", bytes.fromhex("FF FE 63 00 6d 00 64 00")),
    ],
)
def test_apply_utf16_modifier(v: str, encoding: str, expected: bytes):
    assert apply_utf16_modifier(v, encoding=encoding) == expected


def test_windash_generator():
    assert set(windash_generator(" -param-name ")) == {
        " -param-name ",
        " /param-name ",
        " –param-name ",  # noqa: RUF001
        " —param-name ",
        " ―param-name ",
    }


@pytest.mark.parametrize(
    "modifiers,expected",
    [
        (["exists"], type(None)),
        (["exists", "base64"], ValueError),
        (["exists", "re", "base64"], ValueError),
    ],
)
def test_validate_exists_modifier(modifiers: list[str], expected: Any):
    @safe
    def inner():
        return validate_exists_modifier_condition(modifiers)

    result = inner()
    assert isinstance(result._inner_value, expected)


@pytest.mark.parametrize(
    "modifiers,expected",
    [
        (["wide", "base64"], type(None)),
        (["wide", "base64offset"], type(None)),
        (["wide", "base64offset", "contains"], type(None)),
        (["wide"], ValueError),
        (["base64", "wide"], ValueError),
        (["base64offset", "wide"], ValueError),
    ],
)
def test_validate_wide_modifier_order(modifiers: list[str], expected: Any):
    @safe
    def inner():
        return validate_base64_sub_modifier_condition(modifiers, "wide")

    result = inner()
    assert isinstance(result._inner_value, expected)



import pytest

from tests.utils import build_rule


@pytest.mark.parametrize(
    "event,expected",
    [
        ({"x": "a*ba"}, False),
        ({"x": "aba"}, False),
        ({"x": "a?a"}, False),
        ({"x": "a*a"}, True),
    ],
)
def test_escaped_wildcards_with_literal_starts(event: dict, expected: bool):
    rule = build_rule(
        r"""
detection:
  field:
    x: a\*a
  condition: field
"""
    )
    assert rule.match(event) is expected


@pytest.mark.parametrize(
    "event,expected",
    [
        ({"x": "a*ba"}, False),
        ({"x": "aba"}, False),
        ({"x": "a?a"}, True),
        ({"x": "a*a"}, False),
    ],
)
def test_escaped_wildcards_with_literal_question(event: dict, expected: bool):
    rule = build_rule(
        r"""
detection:
  field:
    x: a\?a
  condition: field
    """
    )
    assert rule.match(event) is expected


@pytest.mark.parametrize(
    "event,expected",
    [
        ({"x": "a*ba"}, True),
        ({"x": "aba"}, True),
        ({"x": "a?a"}, True),
        ({"x": "a*a"}, True),
    ],
)
def test_escaped_wildcards_with_star(event: dict, expected: bool):
    rule = build_rule(
        """
detection:
  field:
    x: a*a
  condition: field
"""
    )
    assert rule.match(event) is expected


@pytest.mark.parametrize(
    "event,expected",
    [
        ({"x": "a*ba"}, False),
        ({"x": "aba"}, True),
        ({"x": "a?a"}, True),
        ({"x": "a*a"}, True),
    ],
)
def test_escaped_wildcards_with_question(event: dict, expected: bool):
    rule = build_rule(
        """
detection:
  field:
    x: a?a
  condition: field
    """
    )
    assert rule.match(event) is expected



import base64
import ipaddress
from typing import Any, Callable, Mapping

import regex as re

from azuma import types
from azuma.exceptions import UnsupportedFeatureError

from .utils import replace_placeholders, replace_with_placeholder

# TODO We need to support the rest of them
SUPPORTED_MODIFIERS = {
    "all",
    "base64",
    "base64offset",
    "cased",
    "cidr",
    "contains",
    "endswith",
    "exists",
    "gt",
    "gte",
    "lt",
    "lte",
    "re",
    "startswith",
    "utf16",
    "utf16be",
    "utf16le",
    "wide",
    "windash",
    # re sub-modifiers
    "i",
    "m",
    "s",
    # 'expand',
    # 'fieldref',
}


def apply_base64_modifier(x: str | bytes) -> str:
    if isinstance(x, str):
        x = x.replace("\n", "")
        return base64.b64encode(x.encode()).decode()

    return base64.b64encode(x).decode()


def apply_base64offset_modifier(x: str | bytes) -> str:
    # modified from https://github.com/SigmaHQ/pySigma
    # (https://github.com/SigmaHQ/pySigma/blob/main/sigma/modifiers.py: SigmaBase64OffsetModifier)
    start_offsets = (0, 2, 3)
    end_offsets = (None, -3, -2)
    offsets: list[str] = []

    if isinstance(x, str):
        x = x.replace("\n", "")

        for i in range(3):
            offsets.append(
                base64.b64encode(i * b" " + x.encode())[
                    start_offsets[i] : end_offsets[(len(x) + i) % 3]
                ].decode()
            )
    else:
        for i in range(3):
            offsets.append(
                base64.b64encode(i * b" " + x)[
                    start_offsets[i] : end_offsets[(len(x) + i) % 3]
                ].decode()
            )

    return f"({'|'.join(offsets)})"


WINDASH_PATTERN = re.compile("\\B[-/]\\b")

WINDASH_PLACEHOLDERS = (
    "-",
    "/",
    chr(int("2013", 16)),  # en_dash
    chr(int("2014", 16)),  # em_dash
    chr(int("2015", 16)),  # horizontal_bar
)


def windash_generator(x: str):
    replaced = replace_with_placeholder(x, WINDASH_PATTERN, "_windash")

    for placeholder in WINDASH_PLACEHOLDERS:
        yield replace_placeholders(replaced, placeholder)


def apply_windash_modifier(x: str) -> str:
    modified = set(windash_generator(x))
    return f"({'|'.join(modified)})"


def apply_utf16_modifier(x: str, encoding: str) -> bytes:
    if encoding == "utf-16":
        return bytes.fromhex("FF FE") + apply_utf16_modifier(x, "utf-16le")

    modified = b""
    for c in x:
        modified += c.encode(encoding)

    return modified


MODIFIER_FUNCTIONS: Mapping[str, Callable[[str], Any]] = {
    "base64": lambda x: apply_base64_modifier(x),
    "base64offset": lambda x: apply_base64offset_modifier(x),
    "contains": lambda x: f".*{x}.*",
    "endswith": lambda x: f".*{x}$",
    "startswith": lambda x: f"^{x}.*",
    "utf16": lambda x: apply_utf16_modifier(x, encoding="utf-16"),
    "utf16be": lambda x: apply_utf16_modifier(x, encoding="utf-16be"),
    "utf16le": lambda x: apply_utf16_modifier(x, encoding="utf-16le"),
    "wide": lambda x: apply_utf16_modifier(x, encoding="utf-16le"),
    "windash": lambda x: apply_windash_modifier(x),
}


def process_field_name(field_string: str) -> tuple[str, list[str]]:
    name_and_modifiers = field_string.split("|")
    name = name_and_modifiers.pop(0)
    modifiers = [m for m in name_and_modifiers if m]

    unsupported = set(modifiers) - SUPPORTED_MODIFIERS
    if unsupported:
        raise UnsupportedFeatureError(
            f"Unsupported field modifiers used: {unsupported}"
        )

    return name, modifiers


_NSC = NON_SPECIAL_CHARACTERS = r"[^\\*?]*"
ESCAPED_SPECIAL_CHARACTER = r"(?:\\[*?])"
ESCAPED_OTHER_CHARACTER = r"(?:\\[^*?])"
ESCAPED_WILDCARD_PATTERN = re.compile(
    rf"(?:{_NSC}{ESCAPED_SPECIAL_CHARACTER}*{ESCAPED_OTHER_CHARACTER})*"
)

UPTO_WILDCARD = re.compile(r"^([^\\?*]+|(?:\\[^?*\\])+)+")


def sigma_string_to_regex(original_value: str) -> str:
    value = original_value
    full_content: list[str] = []

    while value:
        # Grab any content up to the first wildcard
        match = UPTO_WILDCARD.match(value)

        if match:
            # The non regex content in the sigma string, may have characters special to regex
            matched = match.group(0)
            full_content.append(re.escape(matched))
            value = value[len(matched) :]
            continue

        if value.startswith("*"):
            full_content.append(".*")
            value = value[1:]
            continue

        if value.startswith("\\*"):
            full_content.append(re.escape("*"))
            value = value[2:]
            continue

        if value.startswith("?"):
            full_content.append(".")
            value = value[1:]
            continue

        if value.startswith("\\?"):
            full_content.append(re.escape("?"))
            value = value[2:]
            continue

        if value.startswith(r"\\*"):
            full_content.append(re.escape("\\") + ".*")
            value = value[3:]
            continue

        if value.startswith(r"\\?"):
            full_content.append(re.escape("\\") + ".")
            value = value[3:]
            continue

        if value.startswith("\\"):
            full_content.append(re.escape("\\"))
            value = value[1:]
            continue

        raise ValueError(f"Could not parse string matching pattern: {original_value}")

    return "".join(full_content)


def get_modified_value(value: str, modifiers: list[str] | None) -> str:
    if not modifiers:
        # If there are no modifiers, we assume exact match
        return f"^{value}$"

    for mod in modifiers:
        func = MODIFIER_FUNCTIONS.get(mod)
        value = func(value) if func else value

    return value


# NOTE: regex flags for non-re modifier case
MODIFIER_REGEX_FLAGS = re.V1 | re.DOTALL


def validate_base64_sub_modifier_condition(
    modifiers: list[str], sub_modifier: str
) -> None:
    has_base64 = "base64" in modifiers
    has_base64offset = "base64offset" in modifiers
    if all([not has_base64, not has_base64offset]):
        raise ValueError(
            f"{sub_modifier} modifier must be used with base64 or base64offset modifier"
        )

    sub_modifier_index = modifiers.index(sub_modifier)
    base64_index = modifiers.index("base64") if has_base64 else None
    base64offset_index = modifiers.index("base64offset") if has_base64offset else None
    if sub_modifier_index > (base64_index or base64offset_index or 0):
        raise ValueError(
            f"{sub_modifier} modifier must be used before base64 or base64offset modifier"
        )


def validate_exists_modifier_condition(modifiers: list[str]) -> None:
    if len(modifiers) > 1:
        raise ValueError("exists modifier cannot use along with other modifiers")


def validate_re_modifier_condition(modifiers: list[str]) -> None:
    re_index = modifiers.index("re")
    if re_index > 0:
        raise ValueError("re modifier must be used before other sub-modifiers")

    has_non_sub_modifiers = len(set(modifiers) - {"re", "i", "m", "s"}) > 0
    if has_non_sub_modifiers:
        raise ValueError(
            "re modifier cannot use along with other modifiers except re sub-modifiers"
        )


def apply_re_modifiers(value: str, modifiers: list[str]) -> types.Query:
    if "re" not in modifiers:
        return None

    validate_re_modifier_condition(modifiers)

    flags = re.V1
    if "i" in modifiers:
        flags |= re.IGNORECASE
    if "m" in modifiers:
        flags |= re.MULTILINE
    if "s" in modifiers:
        flags |= re.DOTALL

    return re.compile(value, flags=flags)


def apply_cidr_modifier(value: str, modifiers: list[str]) -> types.Query:
    if "cidr" not in modifiers:
        return None

    return lambda x: ipaddress.ip_address(x) in ipaddress.ip_network(value)  # type: ignore


def apply_lte_modifier(value: str, modifiers: list[str]) -> types.Query:
    if "lte" not in modifiers:
        return None

    return lambda x: float(x) <= float(value)  # type: ignore


def apply_lt_modifier(value: str, modifiers: list[str]) -> types.Query:
    if "lt" not in modifiers:
        return None

    return lambda x: float(x) < float(value)  # type: ignore


def apply_gte_modifier(value: str, modifiers: list[str]) -> types.Query:
    if "gte" not in modifiers:
        return None

    return lambda x: float(x) >= float(value)  # type: ignore


def apply_gt_modifier(value: str, modifiers: list[str]) -> types.Query:
    if "gt" not in modifiers:
        return None

    return lambda x: float(x) > float(value)  # type: ignore


def apply_modifiers(value: str, modifiers: list[str]) -> types.Query:
    """
    Apply as many modifiers as we can during signature construction
    to speed up the matching stage as much as possible.
    """
    # those modifiers have priority over the rest
    # in other words, they should not be able to use #get_modified_value
    for func in [
        apply_cidr_modifier,
        apply_lte_modifier,
        apply_lt_modifier,
        apply_gte_modifier,
        apply_gt_modifier,
        apply_re_modifiers,
    ]:
        result = func(value, modifiers)
        if result is not None:
            return result

    # re modifier should be applied beforehand if it's set well
    # so there should not be any re sub-modifiers left
    has_re_sub_modifiers = any(sub in modifiers for sub in {"i", "m", "s"})
    if has_re_sub_modifiers:
        raise ValueError("re sub-modifiers must be used with re modifier")

    # base64 sub-modifiers should be used with base64 or base64offset modifier
    base64_sub_modifiers = {"wide", "utf16be", "utf16le"}
    for base64_sub_modifier in base64_sub_modifiers:
        if base64_sub_modifier in modifiers:
            validate_base64_sub_modifier_condition(modifiers, base64_sub_modifier)

    has_cased = "cased" in modifiers
    has_base64 = "base64" in modifiers or "base64offset" in modifiers

    # don't use re.IGNORECASE if cased modifier is used or base64 or base64offset modifier is used
    flags = (
        MODIFIER_REGEX_FLAGS
        if (has_cased or has_base64)
        else MODIFIER_REGEX_FLAGS | re.IGNORECASE
    )

    if not ESCAPED_WILDCARD_PATTERN.fullmatch(value):
        # transform the unescaped wildcards to their regex equivalent
        reg_value = sigma_string_to_regex(value)
        value = get_modified_value(reg_value, modifiers)
        return re.compile(value, flags=flags)

    value = get_modified_value(value, modifiers)
    return str(value).replace("\\*", "*").replace("\\?", "?")


def normalize_field_map(field: dict[str, Any]) -> types.DetectionMap:
    def map_raw_key_value(raw_key: str, value: Any) -> types.DetectionItem:
        key, modifiers = process_field_name(raw_key)

        if value is None:
            return (key, ([None], modifiers))

        has_exists = "exists" in modifiers
        if has_exists:
            validate_exists_modifier_condition(modifiers)

            if value is True:
                # NOTE: use "*" to check whether a field exists or not
                return (key, ([apply_modifiers("*", [])], modifiers))

            if value is False:
                return (key, ([None], modifiers))

            # NOTE: value should not be a list when exists modifier is used
            raise ValueError("exists modifier must be used with boolean value")

        if isinstance(value, list):
            return (
                key,
                (
                    [
                        apply_modifiers(str(v), modifiers) if v is not None else None
                        for v in value
                    ],
                    modifiers,
                ),
            )

        return (key, ([apply_modifiers(str(value), modifiers)], modifiers))

    return [map_raw_key_value(raw_key, value) for raw_key, value in field.items()]



import pytest

from tests.utils import build_rule


@pytest.mark.parametrize(
    "event,expected",
    [
        ({"a": "foo", "b": "bar", "c": "baz"}, True),
        ({"a": "foo"}, True),
        ({"a": "FOO"}, True),
        ({"a": "Foo"}, True),
        ({"a": "bar_foo"}, True),
        ({"a": "foobar"}, False),
        ({"b": "bar"}, False),
        ({"a": "bar", "b": "foo"}, False),
    ],
)
def test_endswith(event: dict, expected: bool):
    rule = build_rule("""
detection:
  foo:
    a|endswith: foo
  condition: foo""")
    assert rule.match(event) is expected



import contextlib
import tempfile
from pathlib import Path

import ci
import pytest
from git.repo import Repo

from azuma import schemas
from azuma.exceptions import UnsupportedFeatureError

# Confirm whether azuma can parse official rules or not


@pytest.fixture(scope="session")
def repo():
    with tempfile.TemporaryDirectory() as dir:
        yield Repo.clone_from("https://github.com/SigmaHQ/sigma", dir)


@pytest.fixture(scope="session")
def paths(repo: Repo):
    dir = Path(repo.working_dir or "")
    return list(dir.glob("./rules/**/*.yml"))


@pytest.mark.skipif(not ci.is_ci(), reason="do fuzzing test in CI")
def test_parse_file(paths: list[Path]):
    for path in paths:
        with contextlib.suppress(UnsupportedFeatureError):
            assert schemas.Rule.model_validate_file(path).title is not None



import pytest

from tests.utils import build_rule


@pytest.mark.parametrize(
    "event,expected",
    [
        ({"a": "cABpAG4AZw"}, True),
        ({"a": "AAaQBuAGcA"}, True),
        ({"a": "wAGkAbgBnA"}, True),
        ({"a": "foo"}, False),
    ],
)
def test_wide(event: dict, expected: bool):
    rule = build_rule(
        """
detection:
  foo:
    a|wide|base64offset: ping
  condition: foo
        """
    )
    assert rule.match(event) is expected



import pytest

from azuma import schemas
from tests.utils import build_rule


@pytest.mark.parametrize(
    "event,expected",
    [
        ({"a": "foo"}, True),
        ({"a": "Foo"}, False),
        ({"a": "foobar"}, False),
    ],
)
def test_re(event: dict, expected: bool):
    rule = build_rule("""
detection:
  selection:
    a|re: ^foo$
  condition: selection
""")
    assert rule.match(event) is expected


@pytest.fixture
def rule_with_i():
    return schemas.Rule.model_validate_yaml(
        """
title: re

logsource:
  category: test
"""
    )


@pytest.mark.parametrize(
    "event,expected",
    [
        ({"a": "foo"}, True),
        ({"a": "Foo"}, True),
        ({"a": "foobar"}, False),
    ],
)
def test_re_with_i(event: dict, expected: bool):
    rule = build_rule("""
detection:
  selection:
    a|re|i: ^foo$
  condition: selection
""")
    assert rule.match(event) is expected


@pytest.mark.parametrize(
    "event,expected",
    [
        ({"a": "A\nB\nX"}, True),
        ({"a": "foobar"}, False),
    ],
)
def test_re_with_m(event: dict, expected: bool):
    rule = build_rule("""
detection:
  selection:
    a|re|m: X
  condition: selection
""")
    assert rule.match(event) is expected


@pytest.mark.parametrize(
    "event,expected",
    [
        ({"a": "foobar"}, True),
        ({"a": "FOOBAR"}, True),
        ({"a": "Foo"}, False),
        ({"a": "Fo\nbar"}, True),
    ],
)
def test_re_with_multi_sub_modifiers(event: dict, expected: bool):
    rule = build_rule("""
detection:
  selection:
    a|re|i|m|s: fo.bar
  condition: selection
""")
    assert rule.match(event) is expected



import pytest

from tests.utils import build_rule


@pytest.mark.parametrize(
    "event,expected",
    [
        (
            {
                "log": "all sorts of red things and blue things were there",
            },
            True,
        ),
    ],
)
def test_substrings(event: dict, expected: bool):
    rule = build_rule(
        """
detection:
  signs:
    - "red things"
    - "blue things"
  condition: signs
    """
    )
    assert rule.match(event) is expected



import pytest

from azuma import schemas
from tests.utils import build_rule


@pytest.fixture
def rule():
    return build_rule(
        """
detection:
  foo:
    a|windash: " -f "
  condition: foo
"""
    )


@pytest.mark.parametrize(
    "event,expected",
    [
        ({"a": " -f "}, True),
        ({"a": " /f "}, True),
        ({"a": "foo"}, False),
    ],
)
def test_windash(event: dict, expected: bool, rule: schemas.Rule):
    assert rule.match(event) is expected



import pytest

from tests.utils import build_rule


@pytest.mark.parametrize(
    "event,expected",
    [
        ({"y": "found"}, True),
        ({"z": "found"}, False),
        ({"y": "found", "x": "also"}, False),
    ],
)
def test_null_and_not_null(event: dict, expected: bool):
    rule = build_rule("""
detection:
  forbid:
    x: null
  filter:
    y: null
  condition: forbid and not filter
""")
    assert rule.match(event) is expected



import pytest

from azuma.validators import is_valid_date_format


@pytest.mark.parametrize(
    "v,expected",
    [
        ("2020-01-01", True),
        ("2020/01/01", False),
        ("2020/01/01 10:00:00", False),
        ("2020-01-01 10:00:00", False),
        ("foo", False),
    ],
)
def test_is_valid_date_format(v: str, expected: bool):
    assert is_valid_date_format(v) is expected



import pytest

from tests.utils import build_rule


@pytest.fixture
def event():
    return {
        "cats": "good",
        "dogs": "good",
        "dog_count": 2,
        "birds": "many",
        "user": {"name": "foo"},
    }


@pytest.fixture
def base_detection():
    return """
detection:
    true_expected: # dogs or dog_count
        - go?d
        - 2
    true_also_expected: # (dogs good or dogs ok) and cats good
        dogs:
            - good
            - ok
        cats: good
    true_cats_expected:
        cats: go*
    true_still_expected: # cats good or birds few
        - cats: good
        - birds: few
    false_expected: # frogs or trees
        - frogs
        - trees
    false_also_expected: # cats good and birds none
        cats: good
        birds: none
"""


@pytest.fixture
def complicated_condition(base_detection: str):
    return (
        base_detection
        + """
    condition: (all of true_*) and (1 of *_expected) and (1 of true_*) and not all of them and (all of them or true_expected)
"""
    )


def test_or_search(event: dict, base_detection: str):
    # Test a signature where the search block is just a list (or operation)
    # Also has an example of the ? wildcard embedded
    rule = build_rule(base_detection + "    condition: true_expected")
    assert rule.match(event) is True


def test_value_or_search(event: dict, base_detection: str):
    # Test a signature where the search block has a list of values (or across those values)
    rule = build_rule(base_detection + "    condition: true_also_expected")
    assert rule.match(event) is True


def test_value_wildcard_search(event: dict, base_detection: str):
    # has an example of the * wildcard embedded
    rule = build_rule(base_detection + "    condition: true_cats_expected")
    assert rule.match(event) is True


def test_and_search(event: dict, base_detection: str):
    # Test a signature where the search block is just a map (and operation)
    rule = build_rule(base_detection + "    condition: true_still_expected")
    assert rule.match(event) is True


def test_complicated_condition(event: dict, complicated_condition: str):
    rule = build_rule(complicated_condition)
    assert rule.match(event) is True
